/*
 * print.c	Routines to print stuff.
 *
 * Version:	@(#)print.c  1.00  26-Oct-1998  miquels@cistron.nl
 *
 */

#include	"autoconf.h"

#include	<stdio.h>
#include	<stdlib.h>
#include	<sys/types.h>
#include	<time.h>
#include	<pwd.h>
#include	<ctype.h>
#include	<string.h>

#include	"libradius.h"

/*
 *	Print a string.
 */
void librad_safeprint(FILE *fp, char *s, int len)
{
	unsigned char *str = (unsigned char *)s;

	if (len < 0) len = strlen(str);

	while (len-- > 0) {
		/*
		 *	Hack: never print trailing zero.
		 *	Some clients send strings with an off-by-one
		 *	length (confused with strings in C).
		 */
		if (len == 0 && *str == 0)
			break;

		if (*str == '\r')
			fputs("\\r", fp);
		else if (*str == '\n')
			fputs("\\n", fp);
		else if (*str == '\t')
			fputs("\\t", fp);
		else if (*str < 32 || (*str >= 128 && *str <= 160))
			fprintf(fp, "\\%03o", *str);
		else
			fputc(*str, fp);
		str++;
	}
}

/*
 *	Print one attribute and value.
 */
void vp_print(FILE *fp, VALUE_PAIR *vp)
{
	DICT_VALUE	*v;
	char		buf[32];
	time_t		t;

	fprintf(fp, "%s = ", vp->name);

	switch (vp->type) {
		case PW_TYPE_STRING:
			fputc('"', fp);
			librad_safeprint(fp, vp->strvalue, vp->length);
			fputc('"', fp);
			break;
		case PW_TYPE_INTEGER:
			if ((v = dict_valbyattr(vp->attribute, vp->lvalue))
			    != NULL)
				fprintf(fp, "%s", v->name);
			else
				fprintf(fp, "%d", vp->lvalue);
			break;
		case PW_TYPE_DATE:
			t = vp->lvalue;
			strftime(buf, sizeof(buf), "%b %e %Y",
				gmtime(&t));
			fprintf(fp, "\"%s\"", buf);
			break;
		case PW_TYPE_IPADDR:
			fputs(ip_ntoa(NULL, vp->lvalue), fp);
			break;
		default:
			fprintf(fp, "\"Unkown type %d, length %d\"",
				vp->type, vp->length);
			break;
	}
}

/*
 *	Print a whole list of attributes, indented by a TAB
 *	and with a newline at the end.
 */
void vp_printlist(FILE *fp, VALUE_PAIR *vp)
{
	for (; vp; vp = vp->next) {
		fprintf(fp, "\t");
		vp_print(fp, vp);
		fprintf(fp, "\n");
	}
}

